"""Performance tests for the Simplenote MCP server."""

import time
from typing import Dict, List, Union
from unittest.mock import MagicMock, patch

import pytest

from simplenote_mcp.server.cache import NoteCache
from simplenote_mcp.server.server import (
    handle_call_tool,
    handle_list_resources,
    handle_read_resource,
    initialize_cache,
)


@pytest.fixture
def mock_large_note_list() -> List[Dict[str, Union[str, List[str]]]]:
    """Create a mock list of many notes for performance testing."""
    return [
        {
            "key": f"note{i}",
            "content": f"Test note {i}\n\nThis is a test note with content for performance testing.",
            "tags": ["test", "performance"] if i % 3 == 0 else ["test"],
            "modifydate": f"2025-04-{i % 30 + 1:02d}T12:00:00Z",
            "createdate": "2025-01-01T00:00:00Z",
        }
        for i in range(1000)
    ]


@pytest.fixture
def setup_performance_cache(
    mock_large_note_list: List[Dict[str, Union[str, List[str]]]],
) -> NoteCache:
    """Set up a NoteCache with a large number of notes for performance testing."""
    mock_client = MagicMock()
    mock_client.get_note_list.return_value = (mock_large_note_list, 0)
    mock_client.get_note.return_value = (mock_large_note_list[0], 0)

    cache = NoteCache(mock_client)

    # Initialize the cache synchronously for testing
    def init_cache():
        # Simulate API delay
        time.sleep(0.1)
        cache._notes = {str(note["key"]): note for note in mock_large_note_list}
        cache._tags = set()
        for note in mock_large_note_list:
            if "tags" in note:
                cache._tags.update(note["tags"])
        cache._initialized = True
        cache._last_sync = time.time()

    init_cache()
    return cache


class TestPerformance:
    """Performance tests for server operations."""

    @pytest.mark.asyncio
    async def test_list_resources_performance(
        self, setup_performance_cache: NoteCache
    ) -> None:
        """Test the performance of listing resources."""
        with (
            patch("simplenote_mcp.server.server.note_cache", setup_performance_cache),
            patch("simplenote_mcp.server.server.get_config") as mock_get_config,
        ):
            # Configure mock config
            mock_config = MagicMock()
            mock_config.default_resource_limit = 100
            mock_get_config.return_value = mock_config

            # Measure performance for different operations

            # 1. Listing without filters
            start_time = time.time()
            resources = await handle_list_resources()
            listing_time = time.time() - start_time
            print(f"Listing 100 resources took {listing_time:.4f} seconds")
            assert listing_time < 0.1, "Listing resources should be fast"
            assert len(resources) == 100

            # 2. Listing with tag filter
            start_time = time.time()
            filtered_resources = await handle_list_resources(tag="performance")
            filter_time = time.time() - start_time
            print(f"Listing resources with tag filter took {filter_time:.4f} seconds")
            assert filter_time < 0.1, "Filtered listing should be fast"
            assert all("performance" in r.meta["tags"] for r in filtered_resources)

            # 3. Listing with larger limit
            start_time = time.time()
            large_resources = await handle_list_resources(limit=500)
            large_listing_time = time.time() - start_time
            print(f"Listing 500 resources took {large_listing_time:.4f} seconds")
            assert large_listing_time < 0.2, (
                "Listing large number of resources should be reasonably fast"
            )
            assert len(large_resources) == 500

    @pytest.mark.asyncio
    async def test_read_resource_performance(
        self, setup_performance_cache: NoteCache
    ) -> None:
        """Test the performance of reading a resource."""
        with (
            patch("simplenote_mcp.server.server.note_cache", setup_performance_cache),
            patch(
                "simplenote_mcp.server.server.get_simplenote_client"
            ) as mock_get_client,
        ):
            # Create a large note for testing read performance
            large_note = {
                "key": "large_note",
                "content": "Large note content\n" + "x" * 100000,  # ~100KB of content
                "tags": ["test", "large"],
                "modifydate": "2025-04-10T12:00:00Z",
                "createdate": "2025-01-01T00:00:00Z",
            }

            # Add large note to cache
            setup_performance_cache._notes["large_note"] = large_note

            # Measure read performance from cache
            start_time = time.time()
            result = await handle_read_resource("simplenote://note/large_note")
            cache_read_time = time.time() - start_time
            print(f"Reading large note from cache took {cache_read_time:.4f} seconds")
            assert cache_read_time < 0.05, "Reading from cache should be very fast"
            assert len(result.contents[0].text) > 100000

            # Simulate API read by removing from cache and setting up mock client
            del setup_performance_cache._notes["large_note"]
            mock_client = MagicMock()
            mock_client.get_note.return_value = (large_note, 0)
            mock_get_client.return_value = mock_client

            # Measure read performance from API
            start_time = time.time()
            result = await handle_read_resource("simplenote://note/large_note")
            api_read_time = time.time() - start_time
            print(f"Reading large note from API took {api_read_time:.4f} seconds")
            assert api_read_time < 0.1, "Reading from API should be reasonably fast"

    @pytest.mark.asyncio
    async def test_search_performance(self, setup_performance_cache: NoteCache) -> None:
        """Test the performance of searching notes."""
        with (
            patch("simplenote_mcp.server.server.note_cache", setup_performance_cache),
            patch(
                "simplenote_mcp.server.server.get_simplenote_client"
            ) as mock_get_client,
        ):
            # Set up mock client
            mock_client = MagicMock()
            mock_get_client.return_value = mock_client

            # Prepare arguments for search
            search_args = {"query": "test", "limit": "50"}

            # Measure search performance
            start_time = time.time()
            result = await handle_call_tool("search_notes", search_args)
            search_time = time.time() - start_time
            print(f"Searching notes took {search_time:.4f} seconds")
            assert search_time < 0.2, "Search should be reasonably fast"

            # Verify search results
            assert len(result) == 1
            import json

            response = json.loads(result[0].text)
            assert response["success"] is True
            assert response["query"] == "test"
            assert len(response["results"]) <= 50

    @pytest.mark.asyncio
    async def test_cache_initialization_performance(self) -> None:
        """Test the performance of initializing the cache."""
        # Create a large mock note list
        large_note_list = [
            {
                "key": f"note{i}",
                "content": f"Test note {i}\n\nThis is content for note {i}.",
                "tags": ["test", "performance"] if i % 3 == 0 else ["test"],
                "modifydate": f"2025-04-{i % 30 + 1:02d}T12:00:00Z",
                "createdate": "2025-01-01T00:00:00Z",
            }
            for i in range(2000)
        ]

        # Mock the Simplenote client and its response
        mock_client = MagicMock()
        mock_client.get_note_list.return_value = (large_note_list, 0)
        # For the second call returning the index mark
        mock_client.get_note_list.side_effect = [
            (large_note_list, 0),
            ({"notes": [], "mark": "test_mark"}, 0),
        ]

        with (
            patch(
                "simplenote_mcp.server.server.get_simplenote_client",
                return_value=mock_client,
            ),
            patch("simplenote_mcp.server.server.note_cache", None),
        ):
            # Measure initialization performance
            start_time = time.time()
            await initialize_cache()
            init_time = time.time() - start_time
            print(f"Initializing cache with 2000 notes took {init_time:.4f} seconds")

            # The time threshold depends on hardware, but should be reasonable
            assert init_time < 2.0, "Cache initialization should be reasonably fast"
